/******************************************************************************
 * CLASSIFICATION: UNLASSIFIED//FOUO>
 *
 * Copyright(C) ProLogic, Incorporated 2008, 1000 Technology Drive, Fairmont, WV 26554
 * Produced pursuant to U.S. Government sponsorship.
 * 
 * Copyright applies to commercial uses only. Not withstanding the above
 * copyright/restriction, the Government possesses DFARS 252.227-7013. 
 * Unlimited Rights if this is technical data and DFARS 252.227-7014. 
 * Unlimited Rights if this is computer software or computer software 
 * documentation.
 * 
 * $Id: $
 *****************************************************************************/
/**
 * @file ConfigImp.h
 * Header file for the libconfig abstract class that defines methods for
 * reading and writing configuration file information.
 */


//////////////////////////////////////////////////////////////////////
// pragmas
//////////////////////////////////////////////////////////////////////

#pragma once


//////////////////////////////////////////////////////////////////////
// includes
//////////////////////////////////////////////////////////////////////

#include <string>
#include <list>
#include <map>
#ifndef UNDER_CE
#include <locale>
#endif


/**
 * This is the abstract class for determining and storing configuration
 * file information.
 *
 * @author:  Dave Webster
 */
class CConfigImp  
{
public:
	/** 
	 * Determine if a section or subkey exists in the hash map table
	 * 
	 * @param The section or the subkey strings to search for.
	 * 
	 * @return true if it exists false if it does not exist
	 */
	virtual bool ExistSection(std::wstring section);

	
	/** 
	 * Determine if a key or value name exists in the hash map table
	 * 
	 * @param strSection The section that the key should be found in
	 * @param strKey The value name to search for.
	 * 
	 * @return true if it exists false if it does not exist
	 */
	virtual bool ExistKey(std::wstring section, std::wstring key);


	/**
	 * Given a section and a key set its value or given a subkey
	 * and a value name set its value data.
	 * 
	 * @param strSection The section std::string or subkey std::string, 
	 * @param strKey The key std::string or value name std::string
	 * @param strValue The value std::string or value data
	 *                 std::string to be set.
	 *
	 * @return 	True if the value was set, else false.
	 */
	virtual bool SetValue(std::wstring section, std::wstring key, std::wstring value);


	/**
	 * Save the information in the table out to a file
	 * 
	 * @param strFilename A std::string containing the file name to save the file.
	 * 
	 * @return:	0 for success and >0 for failure
	 *
	 * @notes This function has only real significance to config file parsing.
	 *        It saves the configuration file output after normalization
	 *        to another file.
	 */
	virtual unsigned long  SaveConfig(std::wstring filename);

	
	/**
	 * Save the information in the table out to the same file read in.
	 *
	 * @return:	0 for success and >0 for failure
	 *
	 */
	unsigned long SaveConfig( );


	/**
	 * Retrieve all the sections or subkeys in the hash table.
	 * 
	 * @return	A wide std::string std::list of section or subkey strings.
	 */
	virtual std::list<std::wstring>  GetSectionNames();

	
	/**
	 * Given a section and a key return its value or given a subkey
	 * and a value name return its value data.
	 *
	 * @param strSection The section std::string or subkey std::string 
	 * @param the key to look up in the section of the configuration
	 *
	 * @return:	The wide std::string value attached to the input key/value name
	 */
	virtual std::wstring        GetValue(std::wstring section, std::wstring key);

	
	/**
	 * Retrieve the std::list of keys associated with a section. Or in
	 * the case of a registry read, all the value names under a
	 * subkey
	 * 
	 * @param strSection The section or subkey wide std::string name
	 * 
	 * @return	A wide std::string std::list of the keys for a given section.
	 */
	virtual std::list<std::wstring>  GetKeys(std::wstring section);

	
	/** 
	 * Parse the configuration source and store the sections/subkeys
	 * and key-value/value name-value data pairs into a hash table 
	 * map.
	 * 
	 * @param strFilename The file name wide std::string or the root key 
	 *                    wide std::string to parse
	 * 
	 * @return	0 for success and >0 for error
	 */
	virtual unsigned long  ReadConfig(std::wstring filename);

	
	/**
	 * Constructor
	 */
	CConfigImp();

	
	/**
	 * Destructor
	 */
	virtual ~CConfigImp();

protected:
	//////////////////////////////////////////////////////////////////////
	// typedefs
	//////////////////////////////////////////////////////////////////////

	typedef std::map<std::wstring,std::wstring> KEY2VALUEMAP;
	typedef std::map<std::wstring,KEY2VALUEMAP*> CONFIGMAP;

	std::wstring    m_szFilename;
	CONFIGMAP m_mapConfig;
	CONFIGMAP::iterator m_mapConfigIterator;
#ifndef UNDER_CE
	void LowerCase(std::string* str, const std::locale* pLoc = NULL);
	void LowerCase(std::wstring* str, const std::locale* pLoc = NULL);
#else
	void LowerCase(std::string* str);
	void LowerCase(std::wstring* str);
#endif
};
