/******************************************************************************
 * CLASSIFICATION: UNLASSIFIED//FOUO>
 *
 * Copyright(C) WebSec Corporation
 *
 *****************************************************************************/
/**
 * @file Subject.h
 * Header file for functions that help to implement "Observer" patterns.
 */

///////////////////////////////////////////////////////////////////////////////
// pragmas
///////////////////////////////////////////////////////////////////////////////

#pragma once


///////////////////////////////////////////////////////////////////////////////
// includes
///////////////////////////////////////////////////////////////////////////////

#include "DereferenceLess.h"
#include "boost/shared_ptr.hpp"
#include <set>
#include <iterator>
#include <stddef.h> //for size_t


/**
* A little wrapper to help implement Observer patterns.
* 
* Wraps a user-specified type (basic or class) with:
* 
* 	a Get() to get the value
* 	a Set() to modify the value and let others know about it
* 	an Update() simply to signal registerd Observers
* 	an Attach() for registering a callback method
* 	a Detach() for de-registering a callback method
* 
* Example:
* 
* 	struct CObserver
* 	{
* 		bool updated;
* 
* 		CObserver() : updated(false) {}
* 
* 		void update( const int& n ) { updated = true; }
* 	};
* 
* 	void CSubjectTest::TestAttach()
* 	{
* 		CSubject<int> nSubject(0);
* 		CObserver observer;
* 
* 		// observer is interested in nSubject
* 		nSubject.Attach( observer, &CObserver::update );
* 
* 		// assert expected initial conditions
* 		CPPUNIT_ASSERT_EQUAL( false, observer.updated );
* 
* 		nSubject.Set( -6 );
* 
* 		// our observers should have been updated by the change
* 		CPPUNIT_ASSERT_EQUAL( true, observer.updated );
* 
* 		// not so interested anymore
* 		nSubject.Detach( observer, &CObserver::update );
* 	}
* 
* This Subject-Observer implementation is about as thin as you can get
* and support multiple Observers.  It uses a set with a simple hash 
* value-based ordering to ensure non-duplicated registration.  The template
* code instantiates an observer functor that wraps the Observer object
* and method pointers, and stores an instance of that functor in the set.
* (Same as anyone would have to do, only templatized.)  The hash value
* is simply the serialized value of the wrapper object, so it uniquely
* identifies an object+method+subject tuple.  This means the same Observer 
* object can be registered multiple times, each time using a different 
* callback method or subject.
*/

namespace //private to this file
{
	///Provides an interface and ordering for type-specific Observer proxies
	struct _CObserverProxy
	{
		virtual ~_CObserverProxy() {}

		virtual void operator()() =0;    //redirect to Observer's callback method

		virtual std::pair<const unsigned char*, const unsigned char*> hash() const =0; //for ordering

		bool operator<( const _CObserverProxy& rhs ) const
		{
			std::pair<const unsigned char*, const unsigned char*> lhsHash = this->hash();
			std::pair<const unsigned char*, const unsigned char*> rhsHash = rhs.hash();

			while( true ) //break out when there's a difference in the strings (below)
			{
				if( lhsHash.first == lhsHash.second ) //reached end if 1st string
				{
					return rhsHash.first != rhsHash.second; //true if lhs shorter than rhs
				}
				else if( rhsHash.first == rhsHash.second ) //reached end of 2nd string
				{
					return false; //lhs longer than rhs
				}
				else if( *lhsHash.first != *rhsHash.first ) //there's a difference
				{
					break; //out of while
				}
				lhsHash.first++; //shift current ptr in each string
				rhsHash.first++; //eventually we'll reach an end or a difference
			}
			return *lhsHash.first < *rhsHash.first;
		}
	};
}



template<typename T>
struct CSubject
{
private:
	///Proxy functor that redirects subject updates to the Observer
	template<class Observer>
	struct CTypedObserverProxy : public _CObserverProxy
	{
		typedef void (Observer::* mem_ptr)(const T&);

		CTypedObserverProxy( const T& subject, Observer& o, mem_ptr pM )
		: m_subject(subject), m_pObserver(&o), m_pMethod(pM)
		{
		}

		void operator()()
		{
			(m_pObserver->*m_pMethod)( m_subject );
		}
		
		std::pair<const unsigned char*, const unsigned char*> hash() const //for ordering
		{
			const unsigned char* pStart = reinterpret_cast<const unsigned char*>(this);
			const unsigned char* pEnd = pStart + sizeof(*this);
			return std::pair<const unsigned char*, const unsigned char*>( pStart, pEnd );
		}

	private:
		const T& m_subject; //subject reference
		Observer* const m_pObserver; //pointer to the Observer
		mem_ptr const m_pMethod;   //pointer to its callback method
	};

	typedef std::set< boost::shared_ptr<_CObserverProxy>, CDereferenceLess> CProxySet;

	T m_value;
	CProxySet m_observers;

public:
	///Ctor
	explicit CSubject( const T& rhs ): m_value(rhs) {}


	///Getter
	const T& Get() const //doesn't throw
	{
		return this->m_value;
	}


	///Sets then updates listeners; returns value for expression nesting
	const T& Set( const T& value ) //doesn't throw
	{
		this->m_value = value;
		this->Update();
		return value;
	}


	///Visits all listeners, invokes their update methods
	void Update() const //doesn't throw
	{
		//make a local copy of the observers so that on an update they can Detach themselves
		//and not screw up this loop: altering a container invalidates any iterators into it.
		const CProxySet setTmp = m_observers;

		for( typename CProxySet::const_iterator p = setTmp.begin(); p != setTmp.end(); ++p )
		{
			try
			{
				(**p)(); //functor redirects to attached listener
			}
			catch(...)
			{
				//keep going
			}
		}
	}


	///Registers a callback method
	template<class Observer>
	void Attach( Observer& observer, void(Observer::*pMethod)(const T&))
	{
		boost::shared_ptr<_CObserverProxy> pTmp( new CTypedObserverProxy<Observer>( this->m_value, observer, pMethod ));
		m_observers.insert( pTmp );
	}


	///UN-registers a callback method
	template<class Observer>
	void Detach( Observer& observer, void(Observer::*pMethod)(const T&))
	{
		boost::shared_ptr<_CObserverProxy> pTempForComparison( new CTypedObserverProxy<Observer>( this->m_value, observer, pMethod ));
		m_observers.erase( pTempForComparison );
	}
};



///Specialization for reference types.
///This is so you can wrap an existing object w/ CSubject facilities w/o requiring 
///that CSubject own that object.  A major difference in implementation is that it's 
///still a reference that gets passed around to Observers but the data being 
///referred to is relocatable, requiring a pointer reference internally.
template<typename T>
struct CSubject<T&>
{
private:
	///Proxy functor that redirects subject updates to the Observer
	template<class Observer>
	struct CTypedObserverProxy : public _CObserverProxy
	{
		typedef void (Observer::* mem_ptr)(const T&);

		CTypedObserverProxy( const T*& subject, Observer& o, mem_ptr pM )
		: m_subject(subject), m_pObserver(&o), m_pMethod(pM)
		{ //require that subject be non-NULL
		}

		void operator()()
		{
			(m_pObserver->*m_pMethod)( *m_subject ); //m_subject not NULL
		}
		
		std::pair<const unsigned char*, const unsigned char*> hash() const //for ordering
		{
			const unsigned char* pStart = reinterpret_cast<const unsigned char*>(this);
			const unsigned char* pEnd = pStart + sizeof(*this);
			return std::pair<const unsigned char*, const unsigned char*>( pStart, pEnd );
		}

	private:
		const T*& m_subject; //reference to pointer to subject
		Observer* const m_pObserver; //pointer to the Observer
		mem_ptr const m_pMethod;   //pointer to its callback method
	};

	typedef std::set< boost::shared_ptr<_CObserverProxy>, CDereferenceLess> CProxySet;

	const T* m_value; //pointer so we can relocate it; never NULL
	CProxySet m_observers;

public:
	///Ctor
	explicit CSubject( const T& rhs ): m_value(&rhs) {} //NOTE: m_value non-NULL


	///Getter
	const T& Get() const //doesn't throw
	{
		return *this->m_value; //m_value never NULL
	}


	///Sets then updates listeners; returns value for expression nesting
	const T& Set( const T& value ) //doesn't throw
	{
		this->m_value = &value ; //NOTE: m_value non-NULL
		this->Update();
		return value;
	}


	///Visits all listeners, invokes their update methods
	void Update() const //doesn't throw
	{
		//make a local copy of the observers so that on an update they can Detach themselves
		//and not screw up this loop: altering a container invalidates any iterators into it.
		const CProxySet setTmp = m_observers;

		for( typename CProxySet::const_iterator p = setTmp.begin(); p != setTmp.end(); ++p )
		{
			try
			{
				(**p)(); //functor redirects to attached listener
			}
			catch(...)
			{
				//keep going
			}
		}
	}


	///Registers a callback method
	template<class Observer>
	void Attach( Observer& observer, void(Observer::*pMethod)(const T&))
	{
		boost::shared_ptr<_CObserverProxy> pTmp( new CTypedObserverProxy<Observer>( this->m_value, observer, pMethod ));
		m_observers.insert( pTmp );
	}


	///UN-registers a callback method
	template<class Observer>
	void Detach( Observer& observer, void(Observer::*pMethod)(const T&))
	{
		boost::shared_ptr<_CObserverProxy> pTempForComparison( new CTypedObserverProxy<Observer>( this->m_value, observer, pMethod ));
		m_observers.erase( pTempForComparison );
	}
};




//specialization for type void ---------------------------------------------------------------------------------------
//basically turns CSubject into a triggering mechanism

template<>
struct CSubject<void>
{
private:
	template<class Observer> //this is the actual proxy object that redirects subject updates to the Observer
	struct CTypedObserverProxy : public _CObserverProxy
	{
		typedef void (Observer::* mem_ptr)();

		CTypedObserverProxy( Observer& o, mem_ptr pM ) : m_pObserver(&o), m_pMethod(pM)
		{
		}

		void operator()()
		{
			(m_pObserver->*m_pMethod)();
		}
		
		std::pair<const unsigned char*, const unsigned char*> hash() const //for ordering
		{
			const unsigned char* pStart = reinterpret_cast<const unsigned char*>(this);
			const unsigned char* pEnd = pStart + sizeof(*this);
			return std::pair<const unsigned char*, const unsigned char*>( pStart, pEnd );
		}

	private:
		Observer* const m_pObserver; //pointer to the Observer
		mem_ptr const m_pMethod;   //pointer to its callback method
	}; //CTypedObserverProxy


	typedef std::set< boost::shared_ptr<_CObserverProxy>, CDereferenceLess> CProxySet;

	CProxySet m_observers;

public:
	//visits all observers, invokes application operator
	void Update() const
	{
		//make a local copy of the observers so that on an update they can Detach themselves
		//and not screw up this loop: altering a container invalidates any iterators into it.
		const CProxySet setTmp = m_observers;

		for( CProxySet::const_iterator p = setTmp.begin(); p != setTmp.end(); ++p )
		{
			try
			{
				(**p)(); //invoke application operator
			}
			catch(...)
			{
				//keep going
			}
		}
	}


	// registers a callback method
	template<class Observer>
	void Attach( Observer& observer, void(Observer::*pMethod)())
	{
		boost::shared_ptr<_CObserverProxy> pTmp( new CTypedObserverProxy<Observer>( observer, pMethod ));
		m_observers.insert( pTmp );
	}


	// UN-registers a callback method
	template<class Observer>
	void Detach( Observer& observer, void(Observer::*pMethod)())
	{
		boost::shared_ptr<_CObserverProxy> pTempForComparison( new CTypedObserverProxy<Observer>( observer, pMethod ));
		m_observers.erase( pTempForComparison );
	}
};



