#include <libusb.h>
#include <stdio.h>
#include <err.h>
#include <stdint.h>

/* <serial port includes> */
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
/* </serial port includes> */

/* after firmware's loaded */
#define LOADED_VENDOR_ID (0x04b4)
#define LOADED_PRODUCT_ID (0x4717)

// Copied from load.c
// TODO move to common file.
static libusb_device *_ref_device_with_vid_pid(
    libusb_context *ctx, uint16_t vendor_id, uint16_t product_id)
{
    struct libusb_device **devs;
    struct libusb_device *found = NULL;
    struct libusb_device *dev;
    size_t i = 0;
    int r;

    if (libusb_get_device_list(ctx, &devs) < 0) {
        return NULL;
    }

    while ((dev = devs[i++]) != NULL) {
        struct libusb_device_descriptor desc;
        r = libusb_get_device_descriptor(dev, &desc);
        if (r < 0) {
            break;
        }
        if (desc.idVendor == vendor_id && desc.idProduct == product_id) {
            found = libusb_ref_device(dev);
            break;
        }
    }

    libusb_free_device_list(devs, 1);
    return found;
}

/**
    Writes data in chunks (see ftdi_write_data_set_chunksize()) to the chip

    \param ftdi pointer to ftdi_context
    \param buf Buffer with the data
    \param size Size of the buffer

    \retval -666: USB device unavailable
    \retval <0: error code from usb_bulk_write()
    \retval >0: number of bytes written
*/
static ssize_t _write_data(libusb_device_handle *handle, const void *data,
    ssize_t size)
{
    const int endpoint = 1; //2; /** 0 is for control */
    const unsigned int timeout = 10000;
    const int chunk_size = 4096;

    unsigned char *p = (unsigned char *)data;
    ssize_t offset = 0;
    int transferred;
    int l;

    while (offset < size)
    {
        int write_size = chunk_size;

        if (offset + write_size > size)
            write_size = size - offset;

        if ((l = libusb_bulk_transfer(handle, endpoint, p, write_size,
            &transferred, timeout)) < 0) {
            warn("%d: %s", __LINE__, libusb_strerror(l));
            return -1;
        }

        offset += transferred;
        p += transferred;
    }

    return offset;
}

static int _set_rs232_mode(libusb_device_handle *handle)
{
    const unsigned tick_rate = 100000000; // 100MHz CPU
    unsigned baud = 9600;
    unsigned ticks_per_symbol = tick_rate/baud;
    unsigned tps_b0 = ticks_per_symbol & 255;
    unsigned tps_b1 = (ticks_per_symbol >> 8) & 255;
    unsigned tps_b2 = (ticks_per_symbol >> 16) & 255;
    unsigned tps_b3 = (ticks_per_symbol >> 24) & 255;

    char set_baud[33];
    snprintf(set_baud, 33, "AT\03330%02x\rAT\03331%02x\rAT\03332%02x\r"
        "AT\03333%02x\r", tps_b3, tps_b2, tps_b1, tps_b0);

    // set baud
    if (32 != _write_data(handle, set_baud, 32))
        return -1;

    // 1 stop bit, no parity, 8 data bits
    if (24 != _write_data(handle, "AT\0333602\rAT\0333500\rAT\0333408\r", 24))
        return -1;

    //const char rs232_mode[] = "AT\0330101\r"; // standard modem
    //const char rs232_mode[] = "AT\0330108\r"; // null modem

    // set rs232 null modem mode
    if (8 != _write_data(handle, "AT\0330108\r", 8))
        return -1;

    // set LED red (2F0000)
    if (24 != _write_data(handle, "AT\033202F\rAT\0332100\rAT\0332200\r", 24))
        return -1;

    return 0;
}

// ep 1 out
// ep 8 in

int main()
{
    libusb_context *context = NULL;
    libusb_device *device = NULL;
    libusb_device_handle *handle = NULL;
    int r = -1;
    int l;

//    int fd = open("/dev/ttyACM0", O_RDWR | O_NOCTTY | O_NONBLOCK);
//    write(fd, "AT\03320FF\rAT\0332100\rAT\0332200\r", 24);
//    close(fd);
//    return 0;


    if ((l = libusb_init(&context)))
    {
        warn("%d: %s", __LINE__, libusb_strerror(l));
        goto out;
    }

    device = _ref_device_with_vid_pid(context, LOADED_VENDOR_ID,
        LOADED_PRODUCT_ID);

//#define UNLOADED_VENDOR_ID (0x08a9)
//#define UNLOADED_PRODUCT_ID (0x0302)
//    device = _ref_device_with_vid_pid(context, 0x08a9, 0x0302);

    if (!device)
    {
        warnx("no uphil devices found");
        goto usb_close;
    }

    if ((l = libusb_open(device, &handle)))
    {
        warn("%d: %s", __LINE__, libusb_strerror(l));
        goto usb_close;
    }

    if ((l = libusb_set_auto_detach_kernel_driver(handle, 1))) {
        warn("%d: %s", __LINE__, libusb_strerror(l));
        goto usb_close;
    }

//    // find out if kernel driver is attached
//    int kernel_driver = libusb_kernel_driver_active(handle, 0) == 1;
//
//    if (kernel_driver) {
//        warnx("%d: Kernel driver active", __LINE__);
//        if ((l = libusb_detach_kernel_driver(handle, 0))) //detach it
//            warn("%d: %s", __LINE__, libusb_strerror(l));
//        else
//            warnx("%d: Kernel driver detached", __LINE__);
//    }

    // claim the interface
    if ((l = libusb_claim_interface(handle, 0))) {
        warn("%d: %s", __LINE__, libusb_strerror(l));
        goto usb_close;
    }

//    if ((l = libusb_set_interface_alt_setting(handle, 0, 1))) {
//        warn("%d: %s", __LINE__, libusb_strerror(l));
//        goto usb_release;
//    }

    if (_set_rs232_mode(handle))
    {
        warnx("install failed");
        goto usb_release;
    }

    r = 0;

usb_release:
    libusb_release_interface(handle, 0); // SEGV if NULL

//    if (kernel_driver) {
//        if ((l = libusb_attach_kernel_driver(handle, 0))) //re-attach it
//            warn("%d: %s", __LINE__, libusb_strerror(l));
//        else
//            warnx("%d: Kernel driver attached", __LINE__);
//    }

usb_close:
    libusb_close(handle); // no-op if NULL
    libusb_unref_device(device); // no-op if NULL
    libusb_exit(context); // SEGV if NULL

out:
    return r;
}
