#include "uphil.h"
#include <libusb.h>
#include <stdio.h>
#include <err.h>
#include <stdint.h>

// poorly written header, requires inclusions before it
#include "ezusb.h"

// include generated files containing firmware
#include "gen/bsfuqf.h"
#include "gen/fpgabootloader.h"
#include "gen/usbzxla.h"
#include "gen/VirtualCom.h"

/* NOTES
 *
 * UPhil USB device shows up (lsusb) with idVendor 0x08a9,
 * idProduct 0x0302
 *
 * After firmware gets loaded that device dissapears and we get a 
 * new one, 04b4:4717, and a node /dev/ttyACM0
 *
 * Some actions (like setting mode) require detaching the kernel,
 * while some (like loading firmware) do not.  Re-attaching the 
 * kernel doesn't work, at least not as a normal user.  (I haven't
 * tried as root.)
 */

/* before firmware's loaded */
#define UNLOADED_VENDOR_ID (0x08a9) // USBee
#define UNLOADED_PRODUCT_ID (0x0302) // Uphil

/* after firmware's loaded */
#define LOADED_VENDOR_ID (0x04b4) // Cyprus
#define LOADED_PRODUCT_ID (0x4717) // EZUSB (FX2)

static int _visit(struct libusb_context *context, uint16_t vendor_id, uint16_t
    product_id, int (*visitor)(void *p, struct libusb_device *device), void *p)
{
    struct libusb_device **all;
    struct libusb_device *cur;
    int i, l, r = -1;

    if ((l = libusb_get_device_list(context, &all)) < 0) {
        warnx("%s", libusb_strerror(l));
        goto exit;
    }

    r = 0;
    for (cur = all[(i = 0)]; cur != NULL; cur = all[++i]) {
        struct libusb_device_descriptor desc;
        if ((l = libusb_get_device_descriptor(cur, &desc)) < 0) {
            warnx("%s", libusb_strerror(l));
            continue;
        }
        if (desc.idVendor == vendor_id && desc.idProduct == product_id) {
            if ((r = visitor(p, cur)))
                break;
        }
    }

    libusb_free_device_list(all, 1);

exit:
    return r;
}

int uphil_visit(struct libusb_context *context, int (*visitor)(void *p, struct
    libusb_device *device), void *p)
{
    return _visit(context, LOADED_VENDOR_ID, LOADED_PRODUCT_ID, visitor, p);
}

int uphil_visit_unloaded(struct libusb_context *context, int (*visitor)(
    void *p, struct libusb_device *device), void *p)
{
    return _visit(context, UNLOADED_VENDOR_ID, UNLOADED_PRODUCT_ID, visitor, p);
}

typedef struct bad_t bad_t;

struct bad_t
{
    int bus;
    int address;
    libusb_device *device;
};

static int _bad_match(void *p, libusb_device *device)
{
    bad_t *self = (bad_t *)p;

    if (libusb_get_device_address(device) == self->address &&
        libusb_get_bus_number(device) == self->bus) {
        self->device = libusb_ref_device(device);
        return 1;
    }
    return 0;
}

libusb_device *uphil_ref(libusb_context *context, int bus, int address)
{
    bad_t bob = { .bus = bus, .address = address, .device = NULL };
    uphil_visit(context, _bad_match, &bob);
    return bob.device;
}

static libusb_device *_ref_unloaded(libusb_context *context, int bus, int address)
{
    bad_t alice = { .bus = bus, .address = address, .device = NULL };
    uphil_visit_unloaded(context, _bad_match, &alice);
    return alice.device;
}

static int _load(libusb_device_handle *handle)
{
    int r = -1;
    int l;

    if ((l = libusb_claim_interface(handle, 0))) {
        warn("%d: %s", __LINE__, libusb_strerror(l));
        goto out;
    }

    if ((l = libusb_set_interface_alt_setting(handle, 0, 1))) {
        warn("%d: %s", __LINE__, libusb_strerror(l));
        goto usb_release;
    }

    if (ezusb_load_ram2(handle, data_r2316_fpgabootloader_bin, data_r2316_fpgabootloader_bin_len) ||
        ezusb_load_bulk2(handle, data_r2336_bsfuqf_ufa, data_r2336_bsfuqf_ufa_len) ||
        ezusb_load_bulk2(handle, data_r2300_usbzxla_bix, data_r2300_usbzxla_bix_len) ||
        ezusb_load_ram2(handle, data_r2329_VirtualCom_bix, data_r2329_VirtualCom_bix_len))
    {
        warnx("install failed");
        goto usb_release;
    }

    r = 0;

usb_release:
    libusb_release_interface(handle, 0); // SEGV if NULL
out:
    return r;
}

int uphil_load_firmware(int bus, int address)
{
    libusb_context *context = NULL;
    libusb_device *device = NULL;
    libusb_device_handle *handle = NULL;
    int r = -1;
    int l;

    if ((l = libusb_init(&context))) {
        warn("%d: %s", __LINE__, libusb_strerror(l));
        goto out;
    }

    if (!(device = _ref_unloaded(context, bus, address))) {
        warnx("no uphil devices found");
        goto usb_close;
    }

    if ((l = libusb_open(device, &handle))) {
        warn("%d: %s", __LINE__, libusb_strerror(l));
        goto usb_close;
    }

    if ((l = _load(handle)))
        goto usb_close;

    r = 0;

usb_close:
    libusb_close(handle); // no-op if NULL
    libusb_unref_device(device); // no-op if NULL
    libusb_exit(context); // SEGV if NULL
out:
    return r;
}

int uphil_load_firmware2(struct libusb_device *device)
{
    libusb_device_handle *handle = NULL;
    int r = -1;
    int l;

    if ((l = libusb_open(device, &handle))) {
        warn("%d: %s", __LINE__, libusb_strerror(l));
        goto usb_close;
    }

    if ((l = _load(handle)))
        goto usb_close;

    r = 0;

usb_close:
    libusb_close(handle); // no-op if NULL
    return r;
}
